/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: audio_alsa_out.c,v 1.12 2000/12/12 21:01:13 guenter Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#include <sys/asoundlib.h>
//#include <linux/asound.h>
//#include <linux/asequencer.h>
//#include <linux/asoundid.h>

#include <inttypes.h>
#include "xine.h"
#include "monitor.h"
#include "audio_out.h"
#include "metronom.h"
#include "ac3.h"

#define AUDIO_NUM_FRAGMENTS    15
#define AUDIO_FRAGMENT_SIZE  8192

extern uint32_t xine_debug;

static uint32_t sample_rate, sample_rate_orig;
static uint32_t num_channels;
static uint32_t bytes_in_buffer;

static uint32_t lastPTS;
static int audio_step; /* pts per 32 768 samples (sample = byte/2) */

static snd_pcm_t *front_handle;
static int pcm_default_card;
static int pcm_default_device;
static int mode = SND_PCM_MODE_STREAM;
static int channel = SND_PCM_CHANNEL_PLAYBACK;

/* 
 * pts <-> samplecount assoc 
 */ 

#define PTS_TAB_NUM 50
static uint32_t audio_pts_tab[PTS_TAB_NUM]; /* organization: ringbuffer */
static uint32_t audio_pts_pos[PTS_TAB_NUM];
static int audio_pts_num, audio_pts_start ;

/* ------------------------------------------------------------------------- */
/*
 * open the audio device for writing to
 */
static uint32_t ao_open(uint32_t bits, uint32_t rate, uint32_t channels) {
  int subdevice = 0;
  int direction = SND_PCM_OPEN_PLAYBACK;  
  snd_pcm_format_t pcm_format;
  snd_pcm_channel_setup_t pcm_chan_setup;
  snd_pcm_channel_params_t pcm_chan_params;
  snd_pcm_channel_info_t pcm_chan_info;
  int bfsize;
  int fsize;
  int err;

  xprintf (VERBOSE|AUDIO, "bits = %d, rate = %d, channels = %d\n", bits, rate, channels);

  if(front_handle != NULL) {

    if(rate == sample_rate_orig)
      return 1;

    snd_pcm_close(front_handle);
  }
  
  lastPTS = 0;
  audio_pts_num = 0;
  audio_pts_start = 0;
  sample_rate_orig = rate;

  if((err = snd_pcm_open_subdevice(&front_handle, pcm_default_card, 
				   pcm_default_device, subdevice, direction
				   | SND_PCM_OPEN_NONBLOCK)) < 0){
    //				   )) < 0){
    perr("snd_pcm_open_subdevice() failed: %s\n", snd_strerror(err));
    return 0;
  }

  memset(&pcm_chan_info, 0, sizeof(snd_pcm_channel_info_t));
  if((err = snd_pcm_channel_info(front_handle, &pcm_chan_info)) < 0) {
    perr("snd_pcm_channel_info() failed: %s\n", snd_strerror(err));
    return 0;
  }	

  memset(&pcm_chan_params, 0, sizeof(snd_pcm_channel_params_t));
  memset(&pcm_format, 0, sizeof(snd_pcm_format_t));
  /* set sample size */
  switch(bits) {
  case 8:
    pcm_format.format = SND_PCM_SFMT_S8;
    break;
    
  case 16:
    pcm_format.format = SND_PCM_SFMT_S16;
    break;

  case 24:
    pcm_format.format = SND_PCM_SFMT_S24;
    break;

  case 32:
    pcm_format.format = SND_PCM_SFMT_S32;
    break;

  default:
    perr("sample format %d unsupported\n", bits);
    break;
  }

  xprintf (VERBOSE|AUDIO, "format name = '%s'\n", 
       snd_pcm_get_format_name(pcm_format.format));

  pcm_format.voices = channels;
  pcm_format.rate = rate;
  pcm_format.interleave = 1;

  num_channels = channels;
  xprintf (VERBOSE|AUDIO, "audio channels = %d\n", num_channels);

  xprintf (VERBOSE|AUDIO, "%d samples/sec\n", rate);

  if(rate > pcm_chan_info.max_rate) {
    perr("\n!!!! WARNING: device only seems to support %d "
	 "samples/sec on this channel\n\n", pcm_chan_info.max_rate);
  }

  sample_rate = rate = pcm_format.rate;

  audio_step = (uint32_t) 90000 * (uint32_t) 32768 / (rate * channels);
  xprintf (VERBOSE|AUDIO, "audio_step %d pts per 32768 samples\n", audio_step);
  metronom_set_audio_rate (audio_step);

  bfsize=0 ;
  fsize = AUDIO_FRAGMENT_SIZE;
  while (fsize>0) {
    fsize /=2;
    bfsize++;
    }
  bfsize--;
  bfsize = (AUDIO_NUM_FRAGMENTS << 16) | bfsize ;
  xprintf (VERBOSE|AUDIO, "Audio buffer fragment info : %x\n", bfsize);

  memcpy(&pcm_chan_params.format, &pcm_format, sizeof(snd_pcm_format_t));

  pcm_chan_params.mode = mode;
  pcm_chan_params.channel = channel;
  //  pcm_chan_params.start_mode = SND_PCM_START_FULL;
  pcm_chan_params.start_mode = SND_PCM_START_DATA;
  //  pcm_chan_params.stop_mode = SND_PCM_STOP_STOP;
  pcm_chan_params.stop_mode = SND_PCM_STOP_ROLLOVER;

  //  pcm_chan_params.buf.stream.fill = SND_PCM_FILL_SILENCE;
  //  pcm_chan_params.buf.stream.max_fill = bfsize+1;

  pcm_chan_params.buf.block.frag_size = bfsize;//1024;//tmp//1024;
  pcm_chan_params.buf.block.frags_max = bfsize+1;//-1;
  pcm_chan_params.buf.block.frags_min = 1;

  snd_pcm_channel_flush(front_handle, channel);
  if((err = snd_pcm_channel_params(front_handle, &pcm_chan_params)) <0){
    perr("snd_pcm_channel_params() failed: %s\n", snd_strerror(err));
    return 0;
  } 
  if((err = snd_pcm_channel_prepare(front_handle, channel)) < 0){
    perr("snd_pcm_channel_prepare() failed: %s\n", snd_strerror(err));
    return 0;
  }

  pcm_chan_setup.mode = mode;
  pcm_chan_setup.channel = channel;

  if((err = snd_pcm_channel_setup(front_handle, &pcm_chan_setup))<0){
    perr("snd_pcm_channel_setup() failed: %s\n", snd_strerror(err));
    return 0;
  }

  printf ("actual rate: %d\n", pcm_chan_setup.format.rate);
  ac3_set_output_rate (pcm_chan_setup.format.rate);

  bytes_in_buffer = 0;

  return 1;
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
static void ao_put_samples(int16_t* output_samples, 
			   uint32_t num_samples, uint32_t pts_) {
  uint32_t vpts;
  snd_pcm_channel_status_t status_front;
  int err;
  
  xprintf(VERBOSE|AUDIO, "Audio : play %d samples at pts %d pos %d \n",
	  num_samples, pts_, bytes_in_buffer);   

  snd_pcm_write(front_handle, output_samples, num_samples<<1);

  memset(&status_front, 0, sizeof(snd_pcm_channel_status_t));
  if((err = snd_pcm_channel_status(front_handle, &status_front)) < 0){
    perr("snd_pcm_channel_status() failed: %s\n", snd_strerror(err));
  }


  /* Hummm, this seems made mistakes (flushing isnt good here). */
  /*
  if(status_front.underrun) {
    perr("underrun, resetting front channel\n");
    snd_pcm_channel_flush(front_handle, channel);
    snd_pcm_playback_prepare(front_handle);
    snd_pcm_write(front_handle, output_samples, num_samples<<1);
    if((err = snd_pcm_channel_status(front_handle, &status_front)) < 0) {
      perr("snd_pcm_channel_status() failed: %s", snd_strerror(err));
    }
    if(status_front.underrun) {
      perr("front write error, giving up\n");
    }
  }
  */
  
  /*
   * remember vpts
   */
  vpts = metronom_got_audio_samples (pts_, num_samples);

  if(audio_pts_num >= PTS_TAB_NUM) {
    audio_pts_tab[audio_pts_start] = vpts;
    audio_pts_pos[audio_pts_start] = bytes_in_buffer;
    
    audio_pts_start = (audio_pts_start + 1) % PTS_TAB_NUM;
  } 
  else {
    audio_pts_tab[audio_pts_num] = vpts;
    audio_pts_pos[audio_pts_num] = bytes_in_buffer;
    audio_pts_num++;
  }

  /*
   * step values
   */
  bytes_in_buffer += num_samples<<1;

}
/* ------------------------------------------------------------------------- */
/*
 *
 */
static uint32_t ao_get_current_pts (void) {
  int pos;
  int i;
  uint32_t ref_pos1, ref_pos2;
  uint32_t ref_pts1, ref_pts2;
  uint32_t m,pts=0;
  snd_pcm_channel_status_t pcm_stat;
  int err;
      
  memset(&pcm_stat, 0, sizeof(snd_pcm_channel_status_t));
  pcm_stat.channel = SND_PCM_CHANNEL_PLAYBACK;
  if((err = snd_pcm_channel_status(front_handle, &pcm_stat)) < 0) {
    //Hide error repport
    //perr("snd_pcm_channel_status() failed: %s\n", snd_strerror(err));
    return 0;
  }
    
  pos = pcm_stat.scount;

  /*
   * tab-based approach
   */
    
  if(audio_pts_num>2) {
    i = 1;
	
    do {
      ref_pos2 = audio_pts_pos[ (audio_pts_start+i)%PTS_TAB_NUM ];
      i++;
    } while ((i<(audio_pts_num)) && (ref_pos2<pos)) ;
    i--;

    ref_pos1 = audio_pts_pos[ (audio_pts_start+i-1)%PTS_TAB_NUM ];

    xprintf(VERBOSE|VPTS, "PTS_TAB: (positions) %d < %d < %d (i=%d)\n",
	   ref_pos1, pos, ref_pos2, i);

    if(ref_pos2<ref_pos1) {

      xprintf (VERBOSE|AUDIO, "clock skew detected!\n");

      i--;
      ref_pos2 = audio_pts_pos[ (audio_pts_start+i)%PTS_TAB_NUM ];
      ref_pos1 = audio_pts_pos[ (audio_pts_start+i-1)%PTS_TAB_NUM ];

      xprintf(VERBOSE|VPTS, "=> PTS_TAB: (positions) %d < %d < %d\n", 
	     ref_pos1, pos, ref_pos2);
    }

    ref_pts2 = audio_pts_tab [ (audio_pts_start+i)%PTS_TAB_NUM ];
    ref_pts1 = audio_pts_tab [ (audio_pts_start+i-1)%PTS_TAB_NUM ];

    xprintf(VERBOSE|VPTS, "PTS_TAB: (pts)       %d < ?? < %d\n", 
	   ref_pts1, ref_pts2);

    /*
     * linear interpolation
     */

    m = (ref_pts2 - ref_pts1) * 32768 / (ref_pos2 - ref_pos1);

    xprintf(VERBOSE|VPTS, "PTS_TAB: m=%d\n",m); 

    pts = ref_pts1 + (pos - ref_pos1) * m / 32768;

    xprintf(VERBOSE|VPTS, "PTS_TAB: => pts = %d (%d sec)\n",
	   pts, pts/90000); 

  }
  else {
    xprintf(VERBOSE|VPTS, "PTS_TAB: Not enough data => 0 pts\n");
  }

  /* 
   * check
   */    
  if(pts < lastPTS) {
    /*
    xprintf(VERBOSE|AUDIO, "Ouch! lastPTS : %d current pts : %d\n",
	    lastPTS, pts);
    */
    printf("Ouch! lastPTS : %d current pts : %d\n",
           lastPTS, pts);
  }

  lastPTS = pts ; 

  return pts;

}
/* ------------------------------------------------------------------------- */
/*
 *
 */
static int ao_get_buf_low (void) {
  snd_pcm_channel_status_t pcm_stat;
  int nFill;
  int err;

  if(front_handle == NULL)
    return 2;

  pcm_stat.channel = SND_PCM_CHANNEL_PLAYBACK;
  if((err = snd_pcm_channel_status(front_handle, &pcm_stat)) < 0) {
    perr("snd_pcm_channel_status() failed: %s\n", snd_strerror(err));
    return 1;
  }
  
  nFill = bytes_in_buffer - pcm_stat.scount ;
  
  if(nFill < (AUDIO_NUM_FRAGMENTS * AUDIO_FRAGMENT_SIZE / 8)) {
    xprintf(VERBOSE|AUDIO, "- Buffer low %d !\n", nFill);
    return 2;
  }
  else if(nFill < (AUDIO_NUM_FRAGMENTS * AUDIO_FRAGMENT_SIZE / 4)) {
    return 1;
  }
  
  return 0; 
}

/* ------------------------------------------------------------------------- */
/*
 *
 */
static void ao_close(void) {
  int err;
  
  if((err = snd_pcm_channel_flush(front_handle, channel)) < 0) {
    perr("snd_pcm_channel_flush() failed: %s\n", snd_strerror(err));
  }

  if((err = snd_pcm_close(front_handle)) < 0) {
    perr("snd_pcm_close() failed: %s\n", snd_strerror(err));
  }

  front_handle = NULL;
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
static ao_functions_t audio_alsaout = {
  ao_open,
  ao_put_samples,
  ao_get_buf_low,
  ao_get_current_pts,
  ao_close
};
/* ------------------------------------------------------------------------- */
/*
 *
 */
ao_functions_t *audio_alsaout_init(void) {
  int best_rate;
  int devnum;
  int err;
  int direction = SND_PCM_OPEN_PLAYBACK;
  int snd_default_card;
  int snd_default_mixer_card;
  int snd_default_mixer_device;
  snd_pcm_info_t pcm_info;
  snd_pcm_channel_info_t pcm_chan_info;
  
  /* Check if, at least, one card is installed */
  if((devnum = snd_cards()) == 0) {
    //    perr("xine cannot live without soundcard at the moment.\n");
    return NULL;
  }
  else {
    snd_default_card = snd_defaults_card();
    if((err = snd_card_load(snd_default_card)) < 0) {
      perr("snd_card_load() failed: %s\n", snd_strerror(err));
    }
    xprintf (VERBOSE|AUDIO, "%d card(s) installed. Default = %d\n", devnum, snd_default_card);

    if((snd_default_mixer_card = snd_defaults_mixer_card()) < 0) {
      perr("snd_defaults_mixer_card() failed: %s\n", 
	   snd_strerror(snd_default_mixer_card));
    }
    xprintf (VERBOSE|AUDIO, "default mixer card = %d\n", snd_default_mixer_card);

    if((snd_default_mixer_device = snd_defaults_mixer_device()) < 0) {
      perr("snd_defaults_mixer_device() failed: %s\n", 
	   snd_strerror(snd_default_mixer_device));
    }
    xprintf (VERBOSE|AUDIO, "default mixer device = %d\n", snd_default_mixer_device);
  }

  xprintf (VERBOSE|AUDIO, "Opening audio device...");
  //    devnum = 0;
  //  best_rate = 0;


  if((pcm_default_card = snd_defaults_pcm_card()) < 0) {
    perr("There is no default pcm card.\n");
    exit(1);
  }
  xprintf (VERBOSE|AUDIO, "snd_defaults_pcm_card() return %d\n", pcm_default_card);
  if((pcm_default_device = snd_defaults_pcm_device()) < 0) {
    perr("There is no default pcm device.\n");
    exit(1);
  }
  xprintf (VERBOSE|AUDIO, "snd_defaults_pcm_device() return %d\n", pcm_default_device);

  if((err = snd_pcm_open(&front_handle, pcm_default_card, 
			 pcm_default_device, direction)) < 0) {
    perr("snd_pcm_open() failed: %s\n", snd_strerror(err));
    exit(1);
  }

  memset(&pcm_info, 0, sizeof(snd_pcm_info_t));
  if((err = snd_pcm_info(front_handle, &pcm_info)) < 0) {
    perr("snd_pcm_info() failed: %s\n", snd_strerror(err));
    exit(1);
  }

  xprintf (VERBOSE|AUDIO, "snd_pcm_info():\n");
  xprintf (VERBOSE|AUDIO, "---------------\n");
  xprintf (VERBOSE|AUDIO, "type      = 0x%x\n", pcm_info.type); 
  xprintf (VERBOSE|AUDIO, "flags     = 0x%x\n", pcm_info.flags);
  xprintf (VERBOSE|AUDIO, "id        = '%s'\n", pcm_info.id);  
  xprintf (VERBOSE|AUDIO, "name      = '%s'\n", pcm_info.name); 
  xprintf (VERBOSE|AUDIO, "playback  = %d\n", pcm_info.playback);
  xprintf (VERBOSE|AUDIO, "capture   = %d\n", pcm_info.capture); 

  memset(&pcm_chan_info, 0, sizeof(snd_pcm_channel_info_t));
  pcm_chan_info.channel = SND_PCM_CHANNEL_PLAYBACK;
  if((err = snd_pcm_channel_info(front_handle, &pcm_chan_info)) < 0) {
    perr("snd_pcm_channel_info() failed: %s\n", snd_strerror(err));
    exit(1);
  }						       

  best_rate = pcm_chan_info.rates;
  xprintf (VERBOSE|AUDIO, "best_rate = %d\n", best_rate);

  xprintf (VERBOSE|AUDIO, "snd_pcm_channel_info(PLAYBACK):\n");
  xprintf (VERBOSE|AUDIO, "-------------------------------\n");
  xprintf (VERBOSE|AUDIO, "subdevice           = %d\n", pcm_chan_info.subdevice);
  xprintf (VERBOSE|AUDIO, "subname             = %s\n", pcm_chan_info.subname);
  xprintf (VERBOSE|AUDIO, "channel             = %d\n", pcm_chan_info.channel);
  xprintf (VERBOSE|AUDIO, "mode                = %d\n", pcm_chan_info.mode);
  xprintf (VERBOSE|AUDIO, "flags               = 0x%x\n", pcm_chan_info.flags);
  xprintf (VERBOSE|AUDIO, "formats             = %d\n", pcm_chan_info.formats);
  xprintf (VERBOSE|AUDIO, "rates               = %d\n", pcm_chan_info.rates);  
  xprintf (VERBOSE|AUDIO, "min_rate            = %d\n", pcm_chan_info.min_rate);
  xprintf (VERBOSE|AUDIO, "max_rate            = %d\n", pcm_chan_info.max_rate);
  xprintf (VERBOSE|AUDIO, "min_voices          = %d\n", pcm_chan_info.min_voices);
  xprintf (VERBOSE|AUDIO, "max_voices          = %d\n", pcm_chan_info.max_voices);
  xprintf (VERBOSE|AUDIO, "buffer_size         = %d\n", pcm_chan_info.buffer_size);
  xprintf (VERBOSE|AUDIO, "min_fragment_size   = %d\n", pcm_chan_info.min_fragment_size);
  xprintf (VERBOSE|AUDIO, "max_fragment_size   = %d\n", pcm_chan_info.max_fragment_size);
  xprintf (VERBOSE|AUDIO, "fragment_align      = %d\n", pcm_chan_info.fragment_align);
  xprintf (VERBOSE|AUDIO, "fifo_size           = %d\n", pcm_chan_info.fifo_size);
  xprintf (VERBOSE|AUDIO, "transfer_block_size = %d\n", pcm_chan_info.transfer_block_size);
  xprintf (VERBOSE|AUDIO, "mmap_size           = %ld\n", pcm_chan_info.mmap_size); 
  xprintf (VERBOSE|AUDIO, "mixer_device        = %d\n", pcm_chan_info.mixer_device);
  
  snd_pcm_close (front_handle);
  front_handle = NULL;

  sample_rate_orig = 0;

  return &audio_alsaout;
}
